//
// Created by genshen(genshenchu@gmail.com) on 2018-03-13.
//

#ifndef KIWI_DATA_DEF_H
#define KIWI_DATA_DEF_H

namespace kiwi {
    typedef unsigned char byte;
    typedef long _type_io_offset;

    typedef int RID; // rank id of MPI.
}

#endif //CRYSTALMD_DATA_DEF_H
